{#-
  This file was automatically generated - do not edit
-#}
{% extends "main.html" %}
{% block tabs %}
  {{ super() }}
  <style>.md-header{position:initial}.md-main__inner{margin:0}.md-content{display:none}@media screen and (min-width:60em){.md-sidebar--secondary{display:none}}@media screen and (min-width:76.25em){.md-sidebar--primary{display:none}}</style>
  <section class="mdx-container">
    <div class="md-grid md-typeset">
      <div class="mdx-hero">
        <div class="mdx-hero__image">
          <img src="assets/images/slice.png" alt="" width="1659" height="1200" draggable="false">
        </div>
        <div class="mdx-hero__content">
          <h1>Build electronic circuit boards from code</h1>
          <p>{{ config.site_description }}</p>
          <a href="{{ page.next_page.url | url }}" title="{{ page.next_page.title | e }}" class="md-button md-button--primary">
            Quick start
          </a>
          <a href="mailto:tim@atopile.io,matt@atopile.io,narayan@atopile.io" title="Email us" class="md-button">
            Contact us
          </a>
        </div>
      </div>
    </div>
  </section>
{% endblock %}
{% block content %}{% endblock %}
{% block footer %}{% endblock %}
